'''
event-consumer

A udata plugin to consume kafka events
'''

import os

from flask import current_app

from .consumer import EventConsumerSingleton
from .commands import consume  # noqa


tag = os.environ.get('CIRCLE_TAG')
build_num = os.environ.get('CIRCLE_BUILD_NUM')

__version__ = '0.1.0.dev' + (str(build_num) if not tag and build_num else '')
__description__ = 'A plugin to consume events'


def init_app(app):
    event_consumer = EventConsumerSingleton.get_instance()

    from .dataset.event import (
        consume_message_resource_analysed,
        consume_message_resource_stored,
        consume_message_resource_checked
    )
    # Register consume functions
    event_consumer.register(
        topics=[f'{current_app.config["UDATA_INSTANCE_NAME"]}.resource.analysed'],
        message_types=['resource.analysed'],
        function=consume_message_resource_analysed
    )
    event_consumer.register(
        topics=[f'{current_app.config["UDATA_INSTANCE_NAME"]}.resource.stored'],
        message_types=['resource.stored'],
        function=consume_message_resource_stored
    )
    event_consumer.register(
        topics=[f'{current_app.config["UDATA_INSTANCE_NAME"]}.resource.checked'],
        message_types=['event-update', 'initialization', 'regular-update'],
        function=consume_message_resource_checked
    )



import logging

from udata.commands import cli

from .consumer import EventConsumerSingleton
log = logging.getLogger(__name__)


@cli.group('event')
def grp():
    '''Event operations'''
    pass


@grp.command()
def consume():
    '''Consume event messages'''
    log.info('Consume event messages')
    event_consumer = EventConsumerSingleton.get_instance()
    event_consumer.consume_kafka_messages()



import logging
from uuid import UUID

from udata.core.dataset.models import get_resource
log = logging.getLogger(__name__)


def consume_message_resource_analysed(key, value):
    '''
    Reads a message and update the resource extras with the payload values.
    '''
    log.info("Consuming message analysed")
    resource = get_resource(UUID(key))
    if resource:
        # TODO: add extra logic here
        for entry in value['value']:
            if entry == 'location' and isinstance(value['value'][entry], dict):
                resource.extras['analysis:location'] = '/'.join([
                    value['value']['location']['url'].strip('/'),
                    value['value']['location']['bucket'],
                    value['value']['location']['key']])
            elif value['value'][entry] is not None:
                resource.extras[f'analysis:{entry}'] = value['value'][entry]
        resource.save()
    else:
        log.warn(f'No resource found for key {key}')


def consume_message_resource_stored(key, value):
    '''
    Reads a message and update the resource extras with the payload values.
    '''
    log.info("Consuming message stored")
    resource = get_resource(UUID(key))
    if resource:
        resource.extras['store:location'] = '/'.join([
            value['value']['location']['netloc'].strip('/'),
            value['value']['location']['bucket'],
            value['value']['location']['key']])
        resource.save()
    else:
        log.warn(f'No resource found for key {key}')


def consume_message_resource_checked(key, value):
    '''
    Reads a message and update the resource extras with the payload values.
    '''
    log.info("Consuming message checked")
    resource = get_resource(UUID(key))
    if resource:
        if 'status' in value['value']:
            resource.extras['check:status'] = value['value']['status']
        if 'timeout' in value['value']:
            resource.extras['check:timeout'] = value['value']['timeout']
        if 'check_date' in value['meta']:
            resource.extras['check:check_date'] = value['meta']['check_date']
        resource.save()
    else:
        log.warn(f'No resource found for key {key}')



import logging

from flask import current_app

from udata_event_service.consumer import consume_kafka


log = logging.getLogger(__name__)


class EventConsumerSingleton():
    registered = {}
    topics = set()

    __instance = None

    @staticmethod
    def get_instance():
        if EventConsumerSingleton.__instance is None:
            EventConsumerSingleton.__instance = EventConsumerSingleton()
        return EventConsumerSingleton.__instance

    def register(self, topics, message_types, function):
        '''Register topics, message types and corresponding callable'''
        if not callable(function):
            msg = 'ExtrasField can only register callables'
            raise TypeError(msg)
        for message_type in message_types:
            self.registered[message_type] = function
        self.topics.update(topics)

    def route_messages(self, key, data, **kwargs):
        try:
            log.debug(f'Consume message {key}')
            message_type = data['meta']['message_type']
            function = self.registered.get(message_type)
            if function:
                function(key, data)
            else:
                log.warn(f'no consumer for the message type {message_type} on this topic')

        except Exception as e:  # Catch and deal with exceptions accordingly
            log.error(f'Error when consuming message {key}: {e}')
            pass

    def consume_kafka_messages(self):
        if self.topics:
            consume_kafka(
                kafka_uri=current_app.config['KAFKA_URI'],
                group_id='udata',
                topics=self.topics,
                message_processing_func=self.route_messages,
            )
        else:
            log.warn('No topics to listen to, exiting.')



import pytest

from udata.core.dataset.factories import DatasetFactory, ResourceFactory

from udata_event_consumer.consumer import EventConsumerSingleton
import udata_event_consumer  # noqa


@pytest.mark.usefixtures('event_consumer_app')
class ConsumerTest:
    def test_consume_resource_stored_event(self):
        dataset = DatasetFactory()
        resource = ResourceFactory()
        dataset.add_resource(resource)

        data = {
            'service': 'datalake',
            'value': {
                'location': {
                    'netloc': 'http://localhost:9000/',
                    'bucket': 'bucket',
                    'key': 'folder/path.csv'
                }
            },
            'meta': {
                'dataset_id': str(dataset.id),
                'message_type': 'resource.stored'
            }
        }

        event_consumer = EventConsumerSingleton.get_instance()
        event_consumer.route_messages(str(resource.id), data)

        dataset.reload()
        resource = dataset.resources[0]
        assert set(resource.extras.keys()) == {'store:location'}
        assert resource.extras['store:location'] == 'http://localhost:9000/bucket/folder/path.csv'

    def test_consume_resource_analysed_csvdetective_event(self):
        dataset = DatasetFactory()
        resource = ResourceFactory()
        dataset.add_resource(resource)

        data = {
            'service': 'csvdetective',
            'value': {
                'location': {
                    'url': 'http://localhost:9000/',
                    'bucket': 'bucket',
                    'key': 'report/path.json'
                }
            },
            'meta': {
                'dataset_id': str(dataset.id),
                'message_type': 'resource.analysed'
            }
        }

        event_consumer = EventConsumerSingleton.get_instance()
        event_consumer.route_messages(str(resource.id), data)

        dataset.reload()
        resource = dataset.resources[0]
        assert list(resource.extras.keys()) == ['analysis:location']
        assert resource.extras['analysis:location'] == 'http://localhost:9000/bucket/report/path.json'

    def test_consume_resource_analysed_hydra_event(self):
        dataset = DatasetFactory()
        resource = ResourceFactory()
        dataset.add_resource(resource)

        data = {
            'service': 'datalake',
            'value': {
                'mime': 'text/plain',
                'resource_url': 'https://static.data.gouv.fr/resources/path.csv',
                'filesize': 154612
            },
            'meta': {
                'dataset_id': str(dataset.id),
                'message_type': 'resource.analysed'
            }
        }

        event_consumer = EventConsumerSingleton.get_instance()
        event_consumer.route_messages(str(resource.id), data)

        dataset.reload()
        resource = dataset.resources[0]
        assert set(resource.extras.keys()) == {'analysis:mime', 'analysis:filesize',
                                               'analysis:resource_url'}
        assert resource.extras['analysis:mime'] == 'text/plain'
        assert resource.extras['analysis:filesize'] == 154612
        assert resource.extras['analysis:resource_url'] == 'https://static.data.gouv.fr/resources/path.csv'

    def test_consume_resource_analysed_hydra_event_with_error(self):
        dataset = DatasetFactory()
        resource = ResourceFactory()
        dataset.add_resource(resource)

        data = {
            'service': 'datalake',
            'value': {
                'resource_url': 'https://static.data.gouv.fr/resources/path.csv',
                'filesize': None,
                'error': 'File too large to download'
            },
            'meta': {
                'dataset_id': str(dataset.id),
                'message_type': 'resource.analysed'
            }
        }

        event_consumer = EventConsumerSingleton.get_instance()
        event_consumer.route_messages(str(resource.id), data)

        dataset.reload()
        resource = dataset.resources[0]
        assert set(resource.extras.keys()) == {'analysis:error', 'analysis:resource_url'}
        assert resource.extras['analysis:resource_url'] == 'https://static.data.gouv.fr/resources/path.csv'
        assert resource.extras['analysis:error'] == 'File too large to download'

    def test_consume_resource_checked(self):
        dataset = DatasetFactory()
        resource = ResourceFactory()
        dataset.add_resource(resource)

        data = {
            'service': 'datalake',
            'value': {
                'url': 'https://www.data.gouv.fr/fr/datasets/r/79b5cac4-4957-486b-bbda-322d80868224',
                'domain': 'www.data.gouv.fr',
                'status': 200,
                'headers': {
                    'server': 'nginx',
                    'date': 'Mon, 02 May 2022 09:58:47 GMT',
                    'content-type': 'text/plain',
                    '...': '...',
                    'access-control-allow-origin': '*',
                    'access-control-allow-methods': 'GET, OPTIONS',
                    'content-encoding': 'gzip'
                },
                'timeout': False,
                'response_time': 0.16390085220336914
            },
            'meta': {
                'dataset_id': str(dataset.id),
                'message_type': 'event-update',
                'check_date': '2020-02-02 20:20:20.202020'
            }
        }

        event_consumer = EventConsumerSingleton.get_instance()
        event_consumer.route_messages(str(resource.id), data)

        dataset.reload()
        resource = dataset.resources[0]
        assert set(resource.extras.keys()) == {'check:check_date', 'check:status',
                                               'check:timeout'}
        assert resource.extras['check:status'] == 200
        assert resource.extras['check:timeout'] is False
        assert resource.extras['check:check_date'] == '2020-02-02 20:20:20.202020'
